#
# Copyright (C) 2021 Associated Universities, Inc. Washington DC, USA.
#
# This file is part of NRAO Workspaces.
#
# Workspaces is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Workspaces is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Workspaces.  If not, see <https://www.gnu.org/licenses/>.
""" Tests for Capability Request """

# pylint: disable=C0301, E0401, E0402, W0611, W0621
from typing import Dict

from hypothesis import given
from hypothesis import strategies as st

from workspaces.capability.enums import CapabilityRequestState, CapabilityVersionState
from workspaces.capability.schema import (
    CapabilityRequest,
    CapabilityVersion,
)
from workspaces.capability.services.capability_info import CapabilityInfo

from ..conftest import (
    SAMPLE_CAPABILITY_NAMES,
    SQLITE_MAX_INT,
    SQLITE_MIN_INT,
    clear_test_database,
)

# Register CapabilityRequest JSON blueprint as a hypothesis type strategy
# To use:
# >>> @given(st.from_type(CapabilityRequest))
# >>> def test(generated_request_json: Dict):
# >>>   request = CapabilityRequest.from_json(generated_request_json)
st.register_type_strategy(
    CapabilityRequest,
    st.fixed_dictionaries(
        {
            "type": st.just("CapabilityRequest"),
            "id": st.integers(min_value=SQLITE_MIN_INT, max_value=SQLITE_MAX_INT),
            "capability_name": st.sampled_from(SAMPLE_CAPABILITY_NAMES),
            "state": st.sampled_from([name for name, _ in CapabilityRequestState.__members__.items()]),
            "ingested": st.booleans(),
            "created_at": st.datetimes().map(
                lambda time: time.isoformat(),
            ),
            "updated_at": st.datetimes().map(
                lambda time: time.isoformat(),
            ),
        }
    ),
)

# Register CapabilityVersion JSON blueprint as a hypothesis type strategy
# To use:
# >>> @given(st.from_type(CapabilityVersion))
# >>> def test(generated_version_json: Dict):
# >>>   version = CapabilityVersion.from_json(generated_version_json)
st.register_type_strategy(
    CapabilityVersion,
    st.fixed_dictionaries(
        {
            "type": st.just("CapabilityVersion"),
            "capability_request_id": st.integers(min_value=SQLITE_MIN_INT, max_value=SQLITE_MAX_INT),
            "version_number": st.integers(min_value=0, max_value=50),
            "parameters": st.one_of(st.lists(st.text()), st.none()),
            "workflow_metadata": st.one_of(
                st.none(), st.dictionaries(st.text(min_size=1), st.text(min_size=1), min_size=1)
            ),
            "files": st.none(),
        }
    ),
)


@given(
    st.from_type(CapabilityRequest),
    st.lists(
        st.from_type(CapabilityVersion), min_size=1, max_size=5, unique_by=lambda version: version["version_number"]
    ),
)
def test_determine_state_complete(mock_capability_info: CapabilityInfo, request_json: Dict, list_of_version_json: Dict):
    """
    Given: A capability request with multiple versions
    When: The current version status is set to Complete
    Then: The request state should be set to Complete
    """
    request = CapabilityRequest.from_json(request_json)
    versions = [CapabilityVersion.from_json(blob) for blob in list_of_version_json]
    current_version_number = max(version.version_number for version in versions)

    with clear_test_database(mock_capability_info):
        mock_capability_info.save_entity(request)

        for version in versions:
            # Associate version with request
            version.capability_request = request
            version.capability_request_id = request.id

            # Set one version to Complete, the rest to Failed or Executing
            if version.version_number == current_version_number:
                version.state = CapabilityVersionState.Complete.name
            else:
                version.state = (
                    CapabilityVersionState.Failed.name
                    if version.state == CapabilityVersionState.Complete.name or not version.state
                    else version.state
                )

            mock_capability_info.save_entity(version)

        request.determine_state()
        assert request.state == CapabilityRequestState.Complete.name


@given(
    st.from_type(CapabilityRequest),
    st.lists(
        st.from_type(CapabilityVersion), min_size=1, max_size=5, unique_by=lambda version: version["version_number"]
    ),
)
def test_determine_state_submitted(
    mock_capability_info: CapabilityInfo, request_json: Dict, list_of_version_json: Dict
):
    """
    Given: A capability request with multiple versions
    When: There are no versions in the Complete state
        And: There is a mix of versions in either the Created, Submitted, or Failed states
    Then: The request state is set to Executing
    """
    request = CapabilityRequest.from_json(request_json)
    versions = [CapabilityVersion.from_json(blob) for blob in list_of_version_json]

    with clear_test_database(mock_capability_info):
        mock_capability_info.save_entity(request)

        for version in versions:
            # Associate version with request
            version.capability_request = request
            version.capability_request_id = request.id

            # Get rid of Complete versions (set their state to Executing instead)
            version.state = (
                CapabilityVersionState.Running.name
                if version.state == CapabilityVersionState.Complete.name or not version.state
                else version.state
            )

            mock_capability_info.save_entity(version)

        request.determine_state()
        assert request.state == CapabilityRequestState.Submitted.name


@given(
    st.from_type(CapabilityRequest),
    st.lists(
        st.from_type(CapabilityVersion), min_size=1, max_size=5, unique_by=lambda version: version["version_number"]
    ),
)
def test_determine_state_failed(mock_capability_info: CapabilityInfo, request_json: Dict, list_of_version_json: Dict):
    """
    Given: A capability request with multiple versions
    When: All versions for the request are in the Failed state
    Then: The request state is set to Failed
    """
    request = CapabilityRequest.from_json(request_json)
    versions = [CapabilityVersion.from_json(blob) for blob in list_of_version_json]

    with clear_test_database(mock_capability_info):
        mock_capability_info.save_entity(request)

        for version in versions:
            # Associate version with request
            version.capability_request = request
            version.capability_request_id = request.id

            # Set all versions to Failed state
            version.state = CapabilityVersionState.Failed.name

            mock_capability_info.save_entity(version)

        request.determine_state()
        assert request.state == CapabilityRequestState.Failed.name


@given(
    st.from_type(CapabilityRequest),
    st.lists(
        st.from_type(CapabilityVersion), min_size=1, max_size=5, unique_by=lambda version: version["version_number"]
    ),
)
def test_determine_state_created(mock_capability_info: CapabilityInfo, request_json: Dict, list_of_version_json: Dict):
    """
    Given: A capability request with multiple versions
    When: All versions for the request are in the Created state
    Then: The request state is set to Created
    """
    request = CapabilityRequest.from_json(request_json)
    versions = [CapabilityVersion.from_json(blob) for blob in list_of_version_json]

    with clear_test_database(mock_capability_info):
        mock_capability_info.save_entity(request)
        for version in versions:
            # Associate version with request
            version.capability_request = request
            version.capability_request_id = request.id

            # Set all versions to Created state
            version.state = CapabilityVersionState.Created.name

            mock_capability_info.save_entity(version)

        request.determine_state()
        assert request.state == CapabilityRequestState.Created.name
